// CubeMapDoc.cpp : implementation file
//

#include "stdafx.h"
#include "DCM.h"
#include "CubeMapDoc.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define FLOOR(x) ((int)(x) - ((x) < 0 && (x) != (int)(x)))
#define CEILING(x) ((int)(x) + ((x) > 0 && (x) != (int)(x)))
#define SQRT2 1.4142135623730950488016887242097f


LPTSTR strCubeDir[CUBEMAP_TOTAL_DIR] = 
{
	_T("Positive X"),
	_T("Negative X"),
	_T("Positive Y"),
	_T("Negative Y"),
	_T("Positive Z"),
	_T("Negative Z")
};

LPTSTR strAppendName[CUBEMAP_TOTAL_DIR] = 
{
	_T("_px"),
	_T("_nx"),
	_T("_py"),
	_T("_ny"),
	_T("_pz"),
	_T("_nz")
};

D3DXVECTOR3 cubeDir[CUBEMAP_TOTAL_DIR] =
{
	D3DXVECTOR3(1, 0, 0),
	D3DXVECTOR3(-1, 0, 0),
	D3DXVECTOR3(0, 1, 0),
	D3DXVECTOR3(0, -1, 0),
	D3DXVECTOR3(0, 0, 1),
	D3DXVECTOR3(0, 0, -1)
};

// if exclude ellipse is present, what faces
// need to be present to prevent holes in cubemap
DWORD cubeDirExcludeMask[CUBEMAP_TOTAL_DIR] = 
{
	(1 << CUBEMAP_POS_Z) | (1 << CUBEMAP_NEG_Z) | (1 << CUBEMAP_POS_Y) | (1 << CUBEMAP_NEG_Y),
	(1 << CUBEMAP_POS_Z) | (1 << CUBEMAP_NEG_Z) | (1 << CUBEMAP_POS_Y) | (1 << CUBEMAP_NEG_Y),

	(1 << CUBEMAP_POS_X) | (1 << CUBEMAP_NEG_X) | (1 << CUBEMAP_POS_Z) | (1 << CUBEMAP_NEG_Z),
	(1 << CUBEMAP_POS_X) | (1 << CUBEMAP_NEG_X) | (1 << CUBEMAP_POS_Z) | (1 << CUBEMAP_NEG_Z),

	(1 << CUBEMAP_POS_X) | (1 << CUBEMAP_NEG_X) | (1 << CUBEMAP_POS_Y) | (1 << CUBEMAP_NEG_Y),
	(1 << CUBEMAP_POS_X) | (1 << CUBEMAP_NEG_X) | (1 << CUBEMAP_POS_Y) | (1 << CUBEMAP_NEG_Y)
};

UINT DialogThread(LPVOID pParam);

/////////////////////////////////////////////////////////////////////////////
// CCubeMapDoc

IMPLEMENT_DYNCREATE(CCubeMapDoc, CDocument)

CCubeMapDoc::CCubeMapDoc()
{
	DWORD i;
	for (i=0; i<CUBEMAP_TOTAL_DIR; i++)
	{
		m_CubeNormals[i] = NULL;
	}
}

CCubeMapDoc::~CCubeMapDoc()
{
	DWORD i;
	if (m_CubeNormals[0] != NULL)
	{
		free(m_CubeNormals[0]);
		for (i=0; i<CUBEMAP_TOTAL_DIR; i++)
		{
			m_CubeNormals[i] = NULL;
		}
	}
}


BOOL CCubeMapDoc::OnNewDocument()
{
	CBMDoc *curBitmapDoc;
	CBMView *curBitmapView;
	CString mess;
	DWORD i;
	CCubeMapSize cmSizeDialog;
	POSITION pos;
	POSITION viewPos;
	HBITMAP hBitmap[CUBEMAP_TOTAL_DIR];
	BITMAPINFO bmi;

	m_CubeMapsFound = 0;
	m_ExclusionPresent = 0;

	// grab all loaded bitmap documents.
	// We want to see if we have enough data
	pos = theApp.m_pBMDocTemplate->GetFirstDocPosition();
	while (pos != NULL)
	{
		curBitmapDoc = DYNAMIC_DOWNCAST(CBMDoc, theApp.m_pBMDocTemplate->GetNextDoc(pos));
		viewPos = curBitmapDoc->GetFirstViewPosition();
		// only 1 view so go ahead and grab it
		curBitmapView = DYNAMIC_DOWNCAST(CBMView, curBitmapDoc->GetNextView(viewPos));

		if (m_CubeMapsFound & (1 << curBitmapDoc->m_CubeDir))
		{
			mess.Format(_T("Duplicate cube map found for %s direction\n"
							"Ignorning this cube map"), strCubeDir[curBitmapDoc->m_CubeDir]);
			MessageBox(NULL, mess, _T("Duplicate Cube Map Direction Error!"), MB_OK);
		}
		else 
		{
			m_CubeIncludeEllipseRects[curBitmapDoc->m_CubeDir] = curBitmapView->GetIncludeEllipseRect();
			m_CubeExcludeEllipseRects[curBitmapDoc->m_CubeDir] = curBitmapView->GetExcludeEllipseRect();

			if (m_CubeIncludeEllipseRects[curBitmapDoc->m_CubeDir]->IsRectEmpty())
			{
				mess.Format(_T("No inclusion rectangle found for %s direction\n"
								"Ignorning this cube map direction"), strCubeDir[curBitmapDoc->m_CubeDir]);
				MessageBox(NULL, mess, _T("No Inclusion Rectangle Error!"), MB_OK);
			}
			else
			{
				m_CubeMapsFound |= (1 << curBitmapDoc->m_CubeDir);
				if (curBitmapView->m_UseExclusion)
				{
					m_ExclusionPresent |= (1 << curBitmapDoc->m_CubeDir);
				}
		
				m_BitmapDocs[curBitmapDoc->m_CubeDir] = curBitmapDoc;
				m_BitmapViews[curBitmapDoc->m_CubeDir] = curBitmapView;
			}
		}

	}

	if (m_CubeMapsFound == 0)
	{
		MessageBox(NULL, _T("No valid included bitmaps found!"), _T("Valid Bitmap Error!"), MB_OK);
		return FALSE;
	}

	for (i=0; i < CUBEMAP_TOTAL_DIR; i++)
	{
		// if there is an exclusion ellipse present
		if (m_ExclusionPresent & (1 << i))
		{
			// make sure that we have other faces to cover excluded ellipse area.
			if (!(m_ExclusionPresent & cubeDirExcludeMask[i]))
			{
				mess.Format(_T("Exclusion ellipse will leave black pixels in cube map.\n"
								"There aren't enough adjacent faces to cover the exclusion \n"
								"range for cube map direction %s.\n\n"
								"Continue with Operation?\n"), strCubeDir[i]);
				if (MessageBox(NULL, mess, _T("Cube map data warning!"), MB_OKCANCEL) == IDCANCEL)
				{
					m_CubeMapsFound = 0;
					return FALSE;
				}
			}
		}
	}
				
	if (cmSizeDialog.DoModal() == IDCANCEL)
	{
		m_CubeMapsFound = 0;
		return FALSE;
	}

	m_CubeSize = cmSizeDialog.m_CubeMapSize;

	bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bmi.bmiHeader.biWidth = m_CubeSize.cx;
	bmi.bmiHeader.biHeight = m_CubeSize.cy;
	bmi.bmiHeader.biPlanes = 1;
	bmi.bmiHeader.biBitCount = 24;
	bmi.bmiHeader.biCompression = BI_RGB;
	bmi.bmiHeader.biSizeImage = m_CubeSize.cx * m_CubeSize.cy;
	bmi.bmiHeader.biXPelsPerMeter = 0;
	bmi.bmiHeader.biYPelsPerMeter = 0;
	bmi.bmiHeader.biClrUsed = 0;
	bmi.bmiHeader.biClrImportant = 0;

	// create the six sides of the cube map now
	for (i=0; i< CUBEMAP_TOTAL_DIR; i++)
	{
		hBitmap[i] = CreateDIBSection(GetDC(NULL), &bmi, DIB_RGB_COLORS, 
								(void**)&m_CubeFacesData[i], NULL, 0);
		if (hBitmap[i] == NULL)
		{
			mess.Format(_T("Unable to create DIB section cube map face %s.\n"),
							strCubeDir[i]);
			MessageBox(NULL, mess, _T("Cube map allocation error!"), MB_OK);
			return FALSE;
		}

	}

	if (!SampleCubeMap())
	{
		return FALSE;
	}

	// create the six sides of the cube map in CBitmap Objects for the view class
	for (i=0; i< CUBEMAP_TOTAL_DIR; i++)
	{
		m_CubeFaces[i].Attach(hBitmap[i]);
	}

	if (!CDocument::OnNewDocument())
	{
		return FALSE;
	}

	SetModifiedFlag();

	UpdateAllViews(NULL);

	return TRUE;
}

BEGIN_MESSAGE_MAP(CCubeMapDoc, CDocument)
	//{{AFX_MSG_MAP(CCubeMapDoc)
	ON_COMMAND(ID_FILE_SAVE, OnFileSave)
	ON_COMMAND(ID_FILE_SAVE_AS, OnFileSaveAs)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CCubeMapDoc diagnostics

#ifdef _DEBUG
void CCubeMapDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CCubeMapDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CCubeMapDoc serialization

void CCubeMapDoc::Serialize(CArchive& ar)
{
	if (ar.IsStoring())
	{
		// TODO: add storing code here
	}
	else
	{
		// TODO: add loading code here
	}
}

/////////////////////////////////////////////////////////////////////////////
// CCubeMapDoc commands


//=====================================================================
// Function:	GetCubeVector
// Description:	returns a 3 vector given a from an x,y, face of a cube
//				map
// Parameters:	face = face of cubemap to calculate vector from
//				tInfo = x,y, width and height of cubemap face texture
//				vecOut = 3 vector pointer for output
// Returns:		pointer to vecOut
//=====================================================================
D3DXVECTOR3 *CCubeMapDoc::GetCubeVector(DWORD face, const CPoint *pos,
									const CSize *size, D3DXVECTOR3 *vecOut)
{
	float s, t, sc, tc;

	// move pixels to center
	s = ((float) pos->x + 0.5f) / (float) size->cx;
	t = ((float) pos->y + 0.5f) / (float) size->cy;
	sc = s*2.0f - 1.0f;
	tc = t*2.0f - 1.0f;

	switch (face) 
	{
	case CUBEMAP_POS_X:
		vecOut->x = 1.0f;
		vecOut->y = -tc;
		vecOut->z = -sc;
		break;
	case CUBEMAP_NEG_X:
		vecOut->x = -1.0f;
		vecOut->y = -tc;
		vecOut->z = sc;
		break;
	case CUBEMAP_POS_Y:
		vecOut->x = sc;
		vecOut->y = 1.0f;
		vecOut->z = tc;
		break;
	case CUBEMAP_NEG_Y:
		vecOut->x = sc;
		vecOut->y = -1.0f;
		vecOut->z = -tc;
		break;
	case CUBEMAP_POS_Z:
		vecOut->x = sc;
		vecOut->y = -tc;
		vecOut->z = 1.0f;
		break;
	case CUBEMAP_NEG_Z: 
		vecOut->x = -sc;
		vecOut->y = -tc;
		vecOut->z = -1.0f;
		break;
	}

	D3DXVec3Normalize(vecOut, vecOut);
	return vecOut;
}

//=====================================================================
// Function:	GetSTCoordinates
// Description:	returns a 2 vector given a from an x,y,z back to a
//				2d texture for the face of a cube
// Parameters:	DWORD face = cube face to map to
//				vecIn = 3 vector for input
//				vecOut = 2 vector pointer for output
// Returns:		2 vector that represents S & T texture coordinates
//=====================================================================
__inline D3DXVECTOR2 *CCubeMapDoc::GetSTCoordinates(DWORD face, 
							const D3DXVECTOR3 *vecIn, D3DXVECTOR2 *vecOut)
{
	D3DXMATRIX mat;
	D3DXVECTOR3 normalVec;

	switch (face) 
	{
	case CUBEMAP_POS_X:
		normalVec.x = sqrtf((vecIn->x + 1.0f) / 2.0f);
		normalVec.y = vecIn->y / (2.0f * normalVec.x);
		normalVec.z = vecIn->z / (2.0f * normalVec.x);
		D3DXVec3Normalize(&normalVec, &normalVec);
		vecOut->x = -normalVec.z;
		vecOut->y = -normalVec.y;
		break;
	case CUBEMAP_NEG_X:
		normalVec.x = sqrtf((1.0f - vecIn->x) / 2.0f);
		normalVec.y = -vecIn->y / (2.0f * normalVec.x);
		normalVec.z = -vecIn->z / (2.0f * normalVec.x);
		D3DXVec3Normalize(&normalVec, &normalVec);
		vecOut->x = normalVec.z;
		vecOut->y = -normalVec.y;
		break;
	case CUBEMAP_POS_Y:
		normalVec.y = sqrtf((vecIn->y + 1.0f) / 2.0f);
		normalVec.x = vecIn->x / (2.0f * normalVec.y);
		normalVec.z = vecIn->z / (2.0f * normalVec.y);
		D3DXVec3Normalize(&normalVec, &normalVec);
		vecOut->x = normalVec.x;
		vecOut->y = normalVec.z;
		break;
	case CUBEMAP_NEG_Y:
		normalVec.y = sqrtf((1.0f - vecIn->y) / 2.0f);
		normalVec.x = -vecIn->x / (2.0f * vecIn->y);
		normalVec.z = -vecIn->z / (2.0f * vecIn->y);
		D3DXVec3Normalize(&normalVec, &normalVec);
		vecOut->x = normalVec.x;
		vecOut->y = -normalVec.z;
		break;
	case CUBEMAP_POS_Z:
		normalVec.z = sqrtf((vecIn->z + 1.0f) / 2.0f);
		normalVec.x = vecIn->x / (2.0f * normalVec.z);
		normalVec.y = vecIn->y / (2.0f * normalVec.z);
		D3DXVec3Normalize(&normalVec, &normalVec);
		vecOut->x = normalVec.x;
		vecOut->y = -normalVec.y;
		break;
	case CUBEMAP_NEG_Z:
		normalVec.z = sqrtf((1.0f - vecIn->z) / 2.0f);
		normalVec.x = -vecIn->x / (2.0f * normalVec.z);
		normalVec.y = -vecIn->y / (2.0f * normalVec.z);
		D3DXVec3Normalize(&normalVec, &normalVec);
		vecOut->x = -normalVec.x;
		vecOut->y = -normalVec.y;
		break;
	}

	return vecOut;
}

//=====================================================================
// Function:	BlendExclusion
// Description:	Blends the exclusion ellipse
// Parameters:	i = cubemap index
//				pos = position in cubemap
//				eBlendRgn = exclusion ellipse -2 pixels region for
//							blending
//				dest = pointer RGB structure to save sample into
// Returns:		.
//=====================================================================
void CCubeMapDoc::BlendExclusion(DWORD index, const CPoint *pos, 
				const CSize *srcSize, const CRgn *eBlendRgn, RGB *dest) 
{
	BOOL blend;
	D3DXVECTOR3 sampleVec;
	D3DXVECTOR3 *pVec;
	D3DXVECTOR2 ST;
	DWORD i;
	int ix, iy;
	float sum[3];
	float divisor;
	float blendWeights[25] = { 1.0f,  1.0f, 1.0f, 
							  1.0f , 1.0f, 1.0f,
							  1.0f,  1.0f, 1.0f 
							};
	int bIndex;

	blend = FALSE;
	// see if need to sample without lookup tables
	if (m_CubeNormals[index] == NULL)
	{
		GetCubeVector(index, pos, &m_CubeSize, &sampleVec);
		pVec = &sampleVec;
	}
	else
	{
		pVec = m_CubeNormals[index] + ((pos->y * m_CubeSize.cx) + pos->x);
	}

	for (i=0; i<CUBEMAP_TOTAL_DIR; i++)
	{
		if (m_CubeMapsFound & (1 << i))
		{
			GetSTCoordinates(i, pVec, &ST);

			ST *= .5f;
			ST += D3DXVECTOR2(.5f, .5f);

			ST.x *= srcSize[i].cx;
			ST.y *= srcSize[i].cy;

			ST -= D3DXVECTOR2(.5f, .5f);

			// clamp coordinates
			ix = min(FLOOR(ST.x), srcSize[i].cx);
			iy = min(FLOOR(ST.y), srcSize[i].cy);
			ix = max(ix, 0);
			iy = max(iy, 0);

			// flip for bitmaps are upside down
			iy = srcSize[i].cy - iy;

			if (eBlendRgn[i].PtInRegion(ix, iy))
			{
				blend = TRUE;
			}
		}
	}

	if (blend)
	{
		sum[0] = 0.0f;
		sum[1] = 0.0f;
		sum[2] = 0.0f;
		divisor = 0.0f;

		dest -= m_CubeSize.cx + 3;
		bIndex = 0;

		for (iy=pos->y-3; iy<=pos->y+3; iy++)
		{
			for (ix=pos->x-3; ix<=pos->x+3; ix++)
			{
				if ((ix >= 0) && (ix < m_CubeSize.cx) &&
						(iy >= 0) && (iy < m_CubeSize.cy))
				{
					sum[0] += (float)dest->R; // * blendWeights[bIndex];
					sum[1] += (float)dest->G; // * blendWeights[bIndex];
					sum[2] += (float)dest->B; // * blendWeights[bIndex];
					divisor += 1.0f; //blendWeights[bIndex];
				}
				dest++;
				bIndex++;
			}
			dest += m_CubeSize.cx - 9;
		}

		dest -= (2 * m_CubeSize.cx)-3;
		dest->R = (UCHAR)(sum[0] / divisor);
		dest->G = (UCHAR)(sum[1] / divisor);
		dest->B = (UCHAR)(sum[2] / divisor);
	}
}
//=====================================================================
// Function:	SampleTextures
// Description:	Samples the individual textures for an ARGB value
// Parameters:	i = cubemap index
//				pos = position in cubemap
//				srcSize = source texture sizes
//				tData = pointer to array of data pointers
//				iRect = inclusion ellipse rectangle
//				eRgn = exclusion ellipse region
//				eBlendRgn = exclusion ellipse -2 pixels region for
//							blending
//				dest = pointer RGB structure to save sample into
// Returns:		.
//=====================================================================
void CCubeMapDoc::SampleTextures(DWORD index, const CPoint *pos, 
					const CSize *srcSize, ILubyte *tData[], 
					const CRect *iRect, const CRgn *eRgn, const CRgn *eBlendRgn,
					RGB *dest) 
{
	D3DXVECTOR3 sampleVec;
	D3DXVECTOR3 *pVec;
	float dotP;
	BYTE *srcPtr;
	D3DXVECTOR2 ST;
	DWORD i;
	int ix, iy;
	float sum[3];
	float divisor;

	float frontThreshold = cosf(D3DXToRadian(45.0f));
	float backThreshold = cosf(D3DXToRadian(90.0f-45.0f));

	// see if need to sample without lookup tables
	if (m_CubeNormals[index] == NULL)
	{
		GetCubeVector(index, pos, &m_CubeSize, &sampleVec);
		pVec = &sampleVec;
	}
	else
	{
		pVec = m_CubeNormals[index] + ((pos->y * m_CubeSize.cx) + pos->x);
	}

	sum[0] = 0.0f;
	sum[1] = 0.0f;
	sum[2] = 0.0f;
	divisor = 0.0f;

	for (i=0; i<CUBEMAP_TOTAL_DIR; i++)
	{
		if (m_CubeMapsFound & (1 << i))
		{
			// find out angle between this cube face and sample direction
			dotP = D3DXVec3Dot(&cubeDir[i], pVec);
			GetSTCoordinates(i, pVec, &ST);

			ST *= .5f;
			ST += D3DXVECTOR2(.5f, .5f);

			ST.x *= srcSize[i].cx;
			ST.y *= srcSize[i].cy;

			ST -= D3DXVECTOR2(.5f, .5f);

			// clamp coordinates
			ix = min(FLOOR(ST.x), srcSize[i].cx);
			iy = min(FLOOR(ST.y), srcSize[i].cy);
			ix = max(ix, 0);
			iy = max(iy, 0);

			// flip for bitmaps are upside down
			iy = srcSize[i].cy - iy;

			// see if it is not in exclusion area
			if (!eRgn[i].PtInRegion(ix, iy))
			{
				// so if it is and it is > than a threshold, use the pixel completely
				if (dotP > frontThreshold)
				{
					dotP = 1.0f;
				}
				else 
				{
					// adjust cos(-backThreshold) to cos(frontThreshold) into 0..1 range
					dotP += backThreshold;		// add cos(backThreshold)
					dotP /= (backThreshold + frontThreshold);			// get scalar
				}

				if (dotP > 0.0f)
				{
					srcPtr = (tData[i] + (iy * (srcSize[i].cx * sizeof(RGB))) +
										(ix * sizeof(RGB)));

					sum[0] += ((float)*srcPtr++) * dotP;
					sum[1] += ((float)*srcPtr++) * dotP;
					sum[2] += ((float)*srcPtr++) * dotP;

					divisor += dotP;
				}

				if (dotP == 1.0f)
				{
					break;
				}
			}
		}
	}

	if (divisor > 0.0f)
	{
		dest->R = (UCHAR)(sum[0] / divisor);
		dest->G = (UCHAR)(sum[1] / divisor);
		dest->B = (UCHAR)(sum[2] / divisor);
	}
	else
	{
		dest->R = 0;
		dest->G = 0;
		dest->B = 0;
	}

}

//=====================================================================
// Function:	SampleCubeMap
// Description:	Samples the cube map from input textures, Uses dot
//				product to figure out weighting from center of bitmap,
//				but doesn't start weighting until dot product is >
//				22.5 degrees
// Parameters:	.
// Returns:		FALSE if error
//=====================================================================
BOOL CCubeMapDoc::SampleCubeMap()
{
	DWORD i;
	BITMAP bi;
	RGB *dest;
	float sx, sy;
	CRgn iRgn[CUBEMAP_TOTAL_DIR];
	CRgn eRgn[CUBEMAP_TOTAL_DIR];
	CRgn eBlendRgn[CUBEMAP_TOTAL_DIR];
	CRect iRect[CUBEMAP_TOTAL_DIR];
	CRect eRect[CUBEMAP_TOTAL_DIR];
	CRect clientRect;
	CRect scaleRect;
	ILuint tImages[CUBEMAP_TOTAL_DIR];
	ILubyte *tData[CUBEMAP_TOTAL_DIR];
	CSize scaledSize[CUBEMAP_TOTAL_DIR];
	int curProgPos;
	CPoint pos;
	int retVal;
	CRect tRect;
	CRgn tempRgn;

	// get data for the found cube source textures
	for (i=0; i<CUBEMAP_TOTAL_DIR; i++)
	{
		if (m_CubeMapsFound & (1 << i))
		{
			// get original bitmap dimensions
			m_BitmapDocs[i]->m_TheBitmap.GetBitmap(&bi);

			// get rectangle in client window coordinates
			scaleRect = *m_BitmapViews[i]->GetIncludeEllipseRect();

			// get Client rectangle
			m_BitmapViews[i]->GetClientRect(&clientRect);
			// scale for client->image
			sx = (float)bi.bmWidth / (float)(clientRect.Width());
			sy = (float)bi.bmHeight / (float)(clientRect.Height());

			// put (scale) rectangle in image area
			scaleRect.left = (DWORD)(scaleRect.left * sx);
			scaleRect.right = (DWORD)(scaleRect.right * sx);
			scaleRect.top = (DWORD)(scaleRect.top * sy);
			scaleRect.bottom = (DWORD)(scaleRect.bottom * sy);

			// generate a cut out image buffer
			ilGenImages(1, &tImages[i]);
			ilBindImage(tImages[i]);
			if (!ilTexImage(scaleRect.Width(), scaleRect.Height(), 1, 3, 
					IL_RGB, IL_UNSIGNED_BYTE, NULL))
			{
				MessageBox(NULL, _T("Unable to allocate memory for image copy.\n"),
						_T("memory error!"), MB_OK);
				return FALSE;
			}

			// get a data pointer
			tData[i]= ilGetData();
			// bind the original image
			ilBindImage(m_BitmapDocs[i]->m_TheImage);
			// copy the pixels to new image
			ilCopyPixels(scaleRect.left, scaleRect.top, 0, scaleRect.Width(), 
				scaleRect.Height(), 1, IL_RGB, IL_UNSIGNED_BYTE, tData[i]);

			// Now rebind new image
			ilBindImage(tImages[i]);

			scaledSize[i].cx = CEILING((float)scaleRect.Width());
			scaledSize[i].cy = CEILING((float)scaleRect.Height());
			// scale image to normalized sphere containing cubemap
//			scaledSize.cx = CEILING((float)m_CubeSize.cx);
//			scaledSize.cy = CEILING((float)m_CubeSize.cy);
//			// enable bilinear filtering
//			iluImageParameter(ILU_FILTER, ILU_BILINEAR);
//			// scale image now
//			iluScale(scaledSize.cx, scaledSize.cy, 1);
			// get a data pointer
			tData[i]= ilGetData();

			iRect[i] = *m_BitmapViews[i]->GetIncludeEllipseRect();
			eRect[i] = *m_BitmapViews[i]->GetExcludeEllipseRect();

			// scale client selected area to new normalized rectangle
			sx = (float)scaledSize[i].cx / (float)clientRect.Width();
			sy = (float)scaledSize[i].cy /(float)clientRect.Height();

			// put (scale) rectangles in image area
			iRect[i].left = (int)((float)iRect[i].left * sx);
			iRect[i].right = (int)((float)iRect[i].right * sx);
			iRect[i].top = (int)((float)iRect[i].top * sy);
			iRect[i].bottom = (int)((float)iRect[i].bottom * sy);

			eRect[i].left = (int)((float)eRect[i].left * sx);
			eRect[i].right = (int)((float)eRect[i].right * sx);
			eRect[i].top = (int)((float)eRect[i].top * sy);
			eRect[i].bottom = (int)((float)eRect[i].bottom * sy);

			eRgn[i].CreateEllipticRgnIndirect(&eRect[i]);

			// scale client selected area to new normalized rectangle
//			sx = (float)m_CubeSize.cx / (float)scaledSize[i].cx;
//			sy = (float)m_CubeSize.cy /(float)scaledSize[i].cy;

//			tRect.left = (int)((float)eRect[i].left * sx);
//			tRect.right = (int)((float)eRect[i].right * sx);
//			tRect.top = (int)((float)eRect[i].top * sy);
//			tRect.bottom = (int)((float)eRect[i].bottom * sy);
			

			tRect = eRect[i];
			tRect.DeflateRect(1, 1);
			tempRgn.CreateEllipticRgnIndirect(&tRect);
			tRect.InflateRect(2, 2);
			eBlendRgn[i].CreateEllipticRgnIndirect(&tRect);
			retVal = eBlendRgn[i].CombineRgn(&eBlendRgn[i], &tempRgn, RGN_DIFF);
			ASSERT(retVal != ERROR);
			tempRgn.DeleteObject();

		}
	}

	// build cubemap normals/reflection vectors for speed
	m_CubeNormals[0] = (D3DXVECTOR3 *)malloc(CUBEMAP_TOTAL_DIR * m_CubeSize.cx * 
								m_CubeSize.cy * sizeof(D3DXVECTOR3));
	if (m_CubeNormals[i] == NULL)
	{
		if (MessageBox(NULL, _T("Couldn't allocate normals of size %d.\nRevert to slow method?"),
				_T("Memory Allocation Error!"), MB_OKCANCEL) == IDCANCEL)
		{
			return FALSE;
		}

		for (i=1; i<CUBEMAP_TOTAL_DIR; i++)
		{
			m_CubeNormals[i] = NULL;
		}

	}
	else
	{
		for (i=1; i<CUBEMAP_TOTAL_DIR; i++)
		{
			m_CubeNormals[i] = m_CubeNormals[0] + (i*(m_CubeSize.cx * m_CubeSize.cy));
		}
	}

	// now loop through each pixel of the cube-map sampling
	// the input textures
	for (i=0; i<CUBEMAP_TOTAL_DIR; i++)
	{
		dest = m_CubeFacesData[i];
		for (pos.y=0; pos.y < m_CubeSize.cy; pos.y++)
		{
			curProgPos = ( ((i+1) * (pos.y+1)) * 10000)
					/ (CUBEMAP_TOTAL_DIR * m_CubeSize.cy);

			for (pos.x = 0; pos.x < m_CubeSize.cx; pos.x++)
			{
				if (m_CubeNormals[i] != NULL)
				{
					// build normal vector table
					GetCubeVector(i, &pos, &m_CubeSize, m_CubeNormals[i] + 
							((pos.y * m_CubeSize.cx) + pos.x));
				}
				SampleTextures(i, &pos, scaledSize, tData, iRect, eRgn, eBlendRgn, dest);
				dest++;
			}
		}
	}

#if 0
	// now loop through each pixel of the cube-map blending exclusion
	// rectangle if it is there
	for (i=0; i<CUBEMAP_TOTAL_DIR; i++)
	{
		dest = m_CubeFacesData[i];
		if ((m_CubeMapsFound & (1 << i)) && (!eRect[i].IsRectEmpty()))
		{
			for (pos.y=0; pos.y < m_CubeSize.cy; pos.y++)
			{
				for (pos.x = 0; pos.x < m_CubeSize.cx; pos.x++)
				{
					BlendExclusion(i, &pos, scaledSize, eBlendRgn, dest);
					dest++;
				}
			}
		}
	}
#endif

	for (i=0; i<CUBEMAP_TOTAL_DIR; i++)
	{
		if (m_CubeMapsFound & (1 << i))
		{
			ilDeleteImages(1, &tImages[i]);
		}
	}


	return TRUE;
}

void CCubeMapDoc::OnFileSave() 
{
	sprintf(SaveFileName, "%s", (LPCSTR)GetPathName());
	if (!GetSaveFileName(&Sfn))
	{
		return;
	}

	OnSaveDocument(Sfn.lpstrFile);
}

BOOL CCubeMapDoc::OnSaveDocument(LPCTSTR lpszPathName) 
{
	DWORD i;
	CString tName;
	CString tExt;
	CString saveName;
	int extPos;

	tName = lpszPathName;
	extPos = tName.ReverseFind('.');
	// strip extension off
	if (extPos == -1)
	{
		tExt = "";
	}
	else
	{
		tExt = tName.Mid(extPos);
		tName = tName.Left(extPos);
	}
	for (i=0; i< CUBEMAP_TOTAL_DIR; i++)
	{
		saveName = tName + strAppendName[i] + tExt;
		if (!ilutWinSaveImage(saveName, m_CubeFaces[i]))
		{
			tName.Format(_T("Unable to save file %s\nError %d"),
					saveName, GetLastError());
			MessageBox(NULL, tName, _T("Save Error!"), MB_OK);
			return FALSE;
		}
	}
	
//	return CDocument::OnSaveDocument(lpszPathName);
	return TRUE;
}

void CCubeMapDoc::OnFileSaveAs() 
{
	strcpy(SaveFileName, "");
	if (!GetSaveFileName(&Sfn))
	{
		return;
	}

	OnSaveDocument(Sfn.lpstrFile);
	
}

BOOL CCubeMapDoc::SaveModified() 
{
	CString message;

	if (!IsModified())
		return TRUE;

	if (!theApp.m_GlobalSkipModified)
	{
		message.Format(_T("Document %s is modified. Do you want to save it?\n"),
					(LPCSTR)GetTitle());
		switch (MessageBox(NULL, message, _T("File not saved warning!"), MB_YESNOCANCEL))
		{
		case IDNO:
			return TRUE;
		case IDCANCEL:
			return FALSE;
		}
	}

	strcpy(SaveFileName, "");
	if (!GetSaveFileName(&Sfn))
	{
		return FALSE;
	}

	return OnSaveDocument(Sfn.lpstrFile);
}
