// CubeMapView.cpp : implementation file
//

#include "stdafx.h"
#include "DCM.h"
#include "CubeMapView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

CRect cubeFaceDestRect[CUBEMAP_TOTAL_DIR] =
{
	CRect(512, 256, 768, 512),		// Positive X
	CRect(0, 256, 256, 512),		// Negative X
	CRect(256, 0, 512, 256),		// Positive Y
	CRect(256, 512, 512, 768),		// Negative Y
	CRect(256, 256, 512, 512),		// Positive Z
	CRect(768, 256, 1024, 512)		// Negative Z
};

/////////////////////////////////////////////////////////////////////////////
// CCubeMapView

IMPLEMENT_DYNCREATE(CCubeMapView, CView)

CCubeMapView::CCubeMapView()
{
	m_ViewReflective = TRUE;
}

CCubeMapView::~CCubeMapView()
{
}


BEGIN_MESSAGE_MAP(CCubeMapView, CView)
	//{{AFX_MSG_MAP(CCubeMapView)
	ON_COMMAND(ID_GENERATE_DCM, OnGenerateDcm)
	ON_UPDATE_COMMAND_UI(ID_GENERATE_DCM, OnUpdateGenerateDcm)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CCubeMapView drawing

void CCubeMapView::OnDraw(CDC* pDC)
{
	BITMAP bitmapInfo;
	CRect cRect;
	CDC memDC;
	CBitmap *oldBitmap;
	CRect destRect;
	DWORD i;

	// Scale image into window for now
	memDC.CreateCompatibleDC(NULL);

	for (i=0; i < CUBEMAP_TOTAL_DIR; i++)
	{

		if (m_ViewReflective)
		{
			CCubeMapDoc* pDoc = DYNAMIC_DOWNCAST(CCubeMapDoc, CView::GetDocument());
			ASSERT_VALID(pDoc);
			oldBitmap = memDC.SelectObject(&pDoc->m_CubeFaces[i]);
			pDoc->m_CubeFaces[i].GetBitmap(&bitmapInfo);
		}
		else
		{
			CDiffCubeMapDoc* pDoc = DYNAMIC_DOWNCAST(CDiffCubeMapDoc, CView::GetDocument());
			ASSERT_VALID(pDoc);
			oldBitmap = memDC.SelectObject(&pDoc->m_CubeFacesDiffuse[i]);
			pDoc->m_CubeFacesDiffuse[i].GetBitmap(&bitmapInfo);
		}

		GetClientRect(cRect);

		float scaleX, scaleY;
		scaleX = (float)cRect.Width() / 1024.0f;
		scaleY = (float)cRect.Height() / 768.0f;

		destRect.left = (DWORD)((float)(cRect.left + cubeFaceDestRect[i].left) * scaleX);
		destRect.right = (DWORD)((float)(cRect.left + cubeFaceDestRect[i].right) * scaleX);
		destRect.top = (DWORD)((float)(cRect.top + cubeFaceDestRect[i].top) * scaleY);
		destRect.bottom = (DWORD)((float)(cRect.top + cubeFaceDestRect[i].bottom) * scaleY);

		pDC->SetStretchBltMode(COLORONCOLOR);
		pDC->StretchBlt(destRect.left, destRect.bottom, destRect.Width(), -destRect.Height(), &memDC,
			0, 0, bitmapInfo.bmWidth, bitmapInfo.bmHeight, SRCCOPY);

		memDC.SelectObject(oldBitmap);
	}

}

/////////////////////////////////////////////////////////////////////////////
// CCubeMapView diagnostics

#ifdef _DEBUG
void CCubeMapView::AssertValid() const
{
	CView::AssertValid();
}

void CCubeMapView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CCubeMapView message handlers

void CCubeMapView::OnInitialUpdate() 
{
	CRect rect;
	float aspect;


	if (m_ViewReflective)
	{
		CCubeMapDoc* pDoc = DYNAMIC_DOWNCAST(CCubeMapDoc, CView::GetDocument());
		aspect = ((float)(pDoc->m_CubeSize.cx * 4) / (float)(pDoc->m_CubeSize.cy * 3));
	}
	else
	{
		CDiffCubeMapDoc* pDoc = DYNAMIC_DOWNCAST(CDiffCubeMapDoc, CView::GetDocument());
		aspect = ((float)(pDoc->m_CubeSize.cx * 4) / (float)(pDoc->m_CubeSize.cy * 3));
	}

	GetClientRect(&rect);

	if (aspect > 1.0f)
	{
		if (((float)rect.Width() * aspect) >= (float)rect.Height())
		{
			rect.right = rect.top + (DWORD)((float)rect.Height() * aspect);
		}
		else
		{
			rect.bottom = rect.top + (DWORD)((float)rect.Width() / aspect);
		}
	}
	else
	{
		if (((float)rect.Height() / aspect) <= (float)rect.Width())
		{
			rect.right = rect.top + (DWORD)((float)rect.Height() * aspect);
		}
		else
		{
			rect.bottom = rect.top + (DWORD)((float)rect.Width() / aspect);
		}
	}

	GetParentFrame()->MoveWindow(rect);

	CView::OnInitialUpdate();
}

void CCubeMapView::OnGenerateDcm() 
{
	CCubeMapDoc *pDoc;
	CCubeMapView *newDCM;
	CRect rcView;
	CDiffCubeMapDoc *pNewDCMDoc;

	pDoc = DYNAMIC_DOWNCAST(CCubeMapDoc, CView::GetDocument());

	pNewDCMDoc = DYNAMIC_DOWNCAST(CDiffCubeMapDoc, theApp.m_pDCMDocTemplate->CreateNewDocument());
	ASSERT_VALID(pNewDCMDoc);

	theApp.m_pDCMDocTemplate->SetDefaultTitle(pNewDCMDoc);
	pNewDCMDoc->SetSourceCM(pDoc);

	CFrameWnd* pFrame = theApp.m_pDCMDocTemplate->CreateNewFrame(pNewDCMDoc, NULL);

	// Create new view using CFrameWnd::CreateView, 
	// which requires a "CCreateContext"
	CCreateContext context;
	context.m_pNewViewClass = RUNTIME_CLASS(CDiffCubeMapView);
	context.m_pCurrentDoc   = pNewDCMDoc;
	newDCM = (CCubeMapView *)pFrame->CreateView(&context);
	ASSERT(newDCM);
	ASSERT_VALID(newDCM);

	if (!pNewDCMDoc->OnNewDocument())
	{
		pFrame->DestroyWindow();
		return;
	}

	pFrame->InitialUpdateFrame(pNewDCMDoc, TRUE);
}

void CCubeMapView::OnUpdateGenerateDcm(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_ViewReflective);
}


