// DCMView.cpp : implementation of the CDCMView class
//

#include "stdafx.h"

#include "DCMDoc.h"
#include "DCMView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define HANDLESIZE 4
/////////////////////////////////////////////////////////////////////////////
// CDCMView

IMPLEMENT_DYNCREATE(CDCMView, CView)

BEGIN_MESSAGE_MAP(CDCMView, CView)
	//{{AFX_MSG_MAP(CDCMView)
	ON_WM_ERASEBKGND()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	//}}AFX_MSG_MAP
	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CView::OnFilePrintPreview)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CDCMView construction/destruction

CDCMView::CDCMView()
{
	CWinApp *theApp;
	m_LbuttonDown = FALSE;
	m_bRectPlaced = FALSE;

	// load in cursors for us

	theApp = AfxGetApp();
	cursors[TOPLEFT] = cursors[BOTTOMRIGHT] = theApp->LoadStandardCursor(IDC_SIZENWSE);
	cursors[TOPMIDDLE] = cursors[BOTTOMMIDDLE] = theApp->LoadStandardCursor(IDC_SIZENS);
	cursors[TOPRIGHT] = cursors[BOTTOMLEFT] = theApp->LoadStandardCursor(IDC_SIZENESW);
	cursors[LEFTMIDDLE] = cursors[RIGHTMIDDLE] = theApp->LoadStandardCursor(IDC_SIZEWE);
	cursors[MOVE] = theApp->LoadStandardCursor(IDC_SIZEALL);
	m_OldCursor = GetCursor();

	m_HandleTracking = UNDEFINEDHANDLE;
}

CDCMView::~CDCMView()
{
}

BOOL CDCMView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CView::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CDCMView drawing

void CDCMView::OnDraw(CDC* pDC)
{
	BITMAP bitmapInfo;
	CRect cRect;
	CDC memDC;
	CBitmap *oldBitmap;

	CDCMDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	// Scale image into window for now
	memDC.CreateCompatibleDC(NULL);
	oldBitmap = memDC.SelectObject(&pDoc->m_TheBitmap);

	pDoc->m_TheBitmap.GetBitmap(&bitmapInfo);
	GetClientRect(cRect);

	pDC->SetStretchBltMode(COLORONCOLOR);
	pDC->StretchBlt(cRect.left, cRect.top, cRect.Width(), cRect.Height(), &memDC,
		0, 0, bitmapInfo.bmWidth, bitmapInfo.bmHeight, SRCCOPY);

	if (m_bRectPlaced)
	{
		DrawSelections();
	}

	memDC.SelectObject(oldBitmap);
}

/////////////////////////////////////////////////////////////////////////////
// CDCMView printing

BOOL CDCMView::OnPreparePrinting(CPrintInfo* pInfo)
{
	// default preparation
	return DoPreparePrinting(pInfo);
}

void CDCMView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add extra initialization before printing
}

void CDCMView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add cleanup after printing
}

/////////////////////////////////////////////////////////////////////////////
// CDCMView diagnostics

#ifdef _DEBUG
void CDCMView::AssertValid() const
{
	CView::AssertValid();
}

void CDCMView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CDCMDoc* CDCMView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CDCMDoc)));
	return (CDCMDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CDCMView message handlers

BOOL CDCMView::OnEraseBkgnd(CDC* pDC) 
{	
	return TRUE;
}

int CDCMView::CheckMouseOverHandles(const CRect &theRect, CPoint mousePos)
{
	CRect temp;
	temp.SetRect(theRect.TopLeft(), theRect.TopLeft());
	temp.InflateRect(CSize(HANDLESIZE, HANDLESIZE));
	if (temp.PtInRect(mousePos))
		return TOPLEFT;

	temp.SetRect(theRect.left + (theRect.Width()/2), theRect.top,
		theRect.left + (theRect.Width()/2), theRect.top);
	temp.InflateRect(CSize(HANDLESIZE, HANDLESIZE));
	if (temp.PtInRect(mousePos))
		return TOPMIDDLE;

	temp.SetRect(theRect.right, theRect.top, theRect.right, theRect.top);
	temp.InflateRect(CSize(HANDLESIZE, HANDLESIZE));
	if (temp.PtInRect(mousePos))
		return TOPRIGHT;

	temp.SetRect(theRect.right, theRect.top + (theRect.Height()/2),
		theRect.right, theRect.top + (theRect.Height()/2));
	temp.InflateRect(CSize(HANDLESIZE,HANDLESIZE));
	if (temp.PtInRect(mousePos))
		return RIGHTMIDDLE;

	temp.SetRect(theRect.BottomRight(), theRect.BottomRight());
	temp.InflateRect(CSize(HANDLESIZE, HANDLESIZE));
	if (temp.PtInRect(mousePos))
		return BOTTOMRIGHT;

	temp.SetRect(theRect.left + (theRect.Width()/2), theRect.bottom,
					theRect.left + (theRect.Width()/2), theRect.bottom);
	temp.InflateRect(CSize(HANDLESIZE,HANDLESIZE));
	if (temp.PtInRect(mousePos))
		return BOTTOMMIDDLE;

	temp.SetRect(theRect.left, theRect.bottom,
					theRect.left, theRect.bottom);
	temp.InflateRect(CSize(HANDLESIZE,HANDLESIZE));
	if (temp.PtInRect(mousePos))
		return BOTTOMLEFT;

	temp.SetRect(theRect.left, theRect.top + (theRect.Height()/2),
		theRect.left, theRect.top + (theRect.Height()/2));
	temp.InflateRect(CSize(HANDLESIZE,HANDLESIZE));
	if (temp.PtInRect(mousePos))
		return LEFTMIDDLE;

	return UNDEFINEDHANDLE;
}

void CDCMView::DrawHandles(CDC *pDC, const CRect &theRect)
{
	CRect temp;

	temp.SetRect(theRect.TopLeft(), theRect.TopLeft());
	temp.InflateRect(CSize(HANDLESIZE, HANDLESIZE));
	pDC->Rectangle(temp);

	temp.SetRect(theRect.left + (theRect.Width()/2), theRect.top,
		theRect.left + (theRect.Width()/2), theRect.top);
	temp.InflateRect(CSize(HANDLESIZE, HANDLESIZE));
	pDC->Rectangle(temp);

	temp.SetRect(theRect.right, theRect.top, theRect.right, theRect.top);
	temp.InflateRect(CSize(HANDLESIZE, HANDLESIZE));
	pDC->Rectangle(temp);

	temp.SetRect(theRect.right, theRect.top + (theRect.Height()/2),
		theRect.right, theRect.top + (theRect.Height()/2));
	temp.InflateRect(CSize(HANDLESIZE,HANDLESIZE));
	pDC->Rectangle(temp);

	temp.SetRect(theRect.BottomRight(), theRect.BottomRight());
	temp.InflateRect(CSize(HANDLESIZE, HANDLESIZE));
	pDC->Rectangle(temp);

	temp.SetRect(theRect.left + (theRect.Width()/2), theRect.bottom,
					theRect.left + (theRect.Width()/2), theRect.bottom);
	temp.InflateRect(CSize(HANDLESIZE,HANDLESIZE));
	pDC->Rectangle(temp);

	temp.SetRect(theRect.left, theRect.bottom,
					theRect.left, theRect.bottom);
	temp.InflateRect(CSize(HANDLESIZE,HANDLESIZE));
	pDC->Rectangle(temp);

	temp.SetRect(theRect.left, theRect.top + (theRect.Height()/2),
		theRect.left, theRect.top + (theRect.Height()/2));
	temp.InflateRect(CSize(HANDLESIZE,HANDLESIZE));
	pDC->Rectangle(temp);

}
void CDCMView::DrawSelections()
{
	POINT polyPoints[5];
	CDC *pDC;
	CPen greenPen(PS_SOLID, 3, RGB(0, 255, 0));
	CPen redPen(PS_SOLID, 3, RGB(255, 0, 0));
	CBrush blackBrush(RGB(0,0,0));
	CPen *oldPen;
	CBrush *oldBrush;
	int oldRop;
	CRect theRect;

	pDC = GetDC();
	oldPen = pDC->SelectObject(&redPen);
	oldBrush = pDC->SelectObject(&blackBrush);

	if (m_bRectPlaced)
	{
		oldRop = pDC->SetROP2(R2_COPYPEN);
	}
	else
	{
		oldRop = pDC->SetROP2(R2_XORPEN);
	}

	polyPoints[0] = m_ExcludeEllipseRect.TopLeft();
	polyPoints[1].x = m_ExcludeEllipseRect.right;
	polyPoints[1].y = m_ExcludeEllipseRect.top;
	polyPoints[2] = m_ExcludeEllipseRect.BottomRight();
	polyPoints[3].x = m_ExcludeEllipseRect.left;
	polyPoints[3].y = m_ExcludeEllipseRect.bottom;
	polyPoints[4] = m_ExcludeEllipseRect.TopLeft();
	pDC->Polyline(polyPoints, 5);

	pDC->Arc(m_ExcludeEllipseRect, m_ExcludeEllipseRect.TopLeft(), 
				m_ExcludeEllipseRect.TopLeft());

	// draw dragging squares
	if (m_bRectPlaced)
	{
		DrawHandles(pDC, m_ExcludeEllipseRect);
	}

	// this is drawn last as it is will be on top for
	// the hittest to match
	pDC->SelectObject(&greenPen);

	polyPoints[0] = m_IncludeEllipseRect.TopLeft();
	polyPoints[1].x = m_IncludeEllipseRect.right;
	polyPoints[1].y = m_IncludeEllipseRect.top;
	polyPoints[2] = m_IncludeEllipseRect.BottomRight();
	polyPoints[3].x = m_IncludeEllipseRect.left;
	polyPoints[3].y = m_IncludeEllipseRect.bottom;
	polyPoints[4] = m_IncludeEllipseRect.TopLeft();
	pDC->Polyline(polyPoints, 5);
	pDC->Arc(m_IncludeEllipseRect, m_IncludeEllipseRect.TopLeft(), m_IncludeEllipseRect.TopLeft());

	// draw dragging squares
	if (m_bRectPlaced)
	{
		DrawHandles(pDC, m_IncludeEllipseRect);
	}

	pDC->SelectObject(oldBrush);
	pDC->SelectObject(oldPen);
	pDC->SetROP2(oldRop);
}

void CDCMView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	int handleIn;
	if (m_bRectPlaced)
	{
		m_bRectPlaced = FALSE;
		RedrawWindow();
	}

	m_LbuttonDown = TRUE;
	handleIn = HitTest(point);
	if (handleIn == UNDEFINEDHANDLE)
	{
		m_IncludeEllipseRect.SetRect(point, point);
		m_ExcludeEllipseRect = m_IncludeEllipseRect;
		m_ExcludeEllipseRect.DeflateRect((m_ExcludeEllipseRect.Width()/8)*3,
				(m_ExcludeEllipseRect.Height()/8)*3);
		m_HandleTracking = UNDEFINEDHANDLE;
	}
	else
	{
		m_HandleTracking = handleIn;
		m_MouseStart = point;
	}

	DrawSelections();
	
	CView::OnLButtonDown(nFlags, point);
}

void CDCMView::OnLButtonUp(UINT nFlags, CPoint point) 
{
	if (m_LbuttonDown)
	{
		// erase selections
		DrawSelections();

		if (!m_IncludeEllipseRect.IsRectEmpty())
		{
			m_bRectPlaced = TRUE;
			// permanently put down the rectangle
			DrawSelections();
		}
	}

	m_LbuttonDown = FALSE;
	
	CView::OnLButtonUp(nFlags, point);
}

int CDCMView::HitTest(CPoint point) 
{
	int handleIn;
	handleIn = CheckMouseOverHandles(m_IncludeEllipseRect, point);
	if (handleIn != UNDEFINEDHANDLE)
	{
		return handleIn;
	}

	handleIn= CheckMouseOverHandles(m_ExcludeEllipseRect, point);
	if (handleIn != UNDEFINEDHANDLE)
	{
		return handleIn | 0x8000;
	}

	// if in include rectangle then move the rectangle
	if (m_IncludeEllipseRect.PtInRect(point))
	{
		m_MouseStart = point;
		return MOVE;
	}

	return UNDEFINEDHANDLE;
}

void CDCMView::DeflateInner(CPoint diff, CPoint mousePos)
{
	CRect newExRect;

	newExRect = m_ExcludeEllipseRect;
	newExRect.DeflateRect(diff.x, diff.y);
	if ((newExRect.left < newExRect.right) && (newExRect.top < newExRect.bottom))
	{
		newExRect.IntersectRect(newExRect, m_IncludeEllipseRect);
		if (newExRect != m_IncludeEllipseRect)
		{
			// erase rectangle
			DrawSelections();
			m_ExcludeEllipseRect = newExRect;
			m_MouseStart = mousePos;
			DrawSelections();
		}
	}
}

void CDCMView::DeflateOuter(CRect diff, CPoint mousePos)
{
	CRect newRect;
	CRect newExRect;

	newRect = m_IncludeEllipseRect;
	newRect.DeflateRect(diff.left, diff.top, diff.right, diff.bottom);

	if ((newRect.left < newRect.right) && (newRect.top < newRect.bottom))
	{
		newExRect = m_ExcludeEllipseRect;
		// center new exclusion ellipse in inclusion ellipse
		newExRect.OffsetRect(newRect.CenterPoint() -
			m_ExcludeEllipseRect.CenterPoint());
		newRect.UnionRect(newRect, m_ExcludeEllipseRect);

		if (newRect != m_ExcludeEllipseRect)
		{
			// erase rectangle
			DrawSelections();
			m_IncludeEllipseRect = newRect;
			m_ExcludeEllipseRect = newExRect;
			m_MouseStart = mousePos;
			DrawSelections();
		}
	}
}

void CDCMView::OnMouseMove(UINT nFlags, CPoint point) 
{
	int handleIn;
	CPoint diff;

	if (m_LbuttonDown && !m_bRectPlaced)
	{

		if (m_HandleTracking == UNDEFINEDHANDLE)
		{
			// xor rectangle off
			DrawSelections();
			m_IncludeEllipseRect.right = point.x;
			m_IncludeEllipseRect.bottom = point.y;
			m_ExcludeEllipseRect = m_IncludeEllipseRect;
			m_ExcludeEllipseRect.DeflateRect((m_IncludeEllipseRect.Width()/8)*3,
				(m_IncludeEllipseRect.Height()/8)*3);
			DrawSelections();
		}
		else
		{
			switch (m_HandleTracking & ~0x8000)
			{
			case MOVE:
				// xor rectangle off
				DrawSelections();
				m_MouseStart = point - m_MouseStart;
				m_IncludeEllipseRect.OffsetRect(m_MouseStart);
				m_ExcludeEllipseRect.OffsetRect(m_MouseStart);
				m_MouseStart = point;
				DrawSelections();
				break;
			case TOPLEFT:
				if (m_HandleTracking & 0x8000)
				{
					diff = point - m_MouseStart;
					DeflateInner(diff, point);
				}
				else
				{
					diff = point - m_MouseStart;
					DeflateOuter(CRect(diff.x, diff.y, 0, 0), point);
				}
				break;
			case TOPMIDDLE:
				if (m_HandleTracking & 0x8000)
				{
					diff = point - m_MouseStart;
					diff.x = 0;
					DeflateInner(diff, point);
				}
				else
				{
					diff = point - m_MouseStart;
					DeflateOuter(CRect(0, diff.y, 0, 0), point);
				}
				break;
			case TOPRIGHT:
				if (m_HandleTracking & 0x8000)
				{
					diff = point - m_MouseStart;
					diff.x = -diff.x;
					DeflateInner(diff, point);
				}
				else
				{
					diff = point - m_MouseStart;
					DeflateOuter(CRect(0, diff.y, -diff.x, 0), point);
				}
				break;
			case RIGHTMIDDLE:
				if (m_HandleTracking & 0x8000)
				{
					diff = point - m_MouseStart;
					diff.x = -diff.x;
					diff.y = 0;
					DeflateInner(diff, point);
				}
				else
				{
					diff = point - m_MouseStart;
					DeflateOuter(CRect(0, 0, -diff.x, 0), point);
				}
				break;
			case BOTTOMRIGHT:
				if (m_HandleTracking & 0x8000)
				{
					diff = -(point - m_MouseStart);
					DeflateInner(diff, point);
				}
				else
				{
					diff = point - m_MouseStart;
					DeflateOuter(CRect(0, 0, -diff.x, -diff.y), point);
				}
				break;
			case BOTTOMMIDDLE:
				if (m_HandleTracking & 0x8000)
				{
					diff = point - m_MouseStart;
					diff.x = 0;
					diff.y = -diff.y;
					DeflateInner(diff, point);
				}
				else
				{
					diff = point - m_MouseStart;
					DeflateOuter(CRect(0, 0, 0, -diff.y), point);
				}
				break;
			case BOTTOMLEFT:
				if (m_HandleTracking & 0x8000)
				{
					diff = point - m_MouseStart;
					diff.y = -diff.y;
					DeflateInner(diff, point);
				}
				else
				{
					diff = point - m_MouseStart;
					DeflateOuter(CRect(diff.x, 0, 0, -diff.y), point);
				}
				break;
			case LEFTMIDDLE:
				if (m_HandleTracking & 0x8000)
				{
					diff = point - m_MouseStart;
					diff.y = 0;
					DeflateInner(diff, point);
				}
				else
				{
					diff = point - m_MouseStart;
					DeflateOuter(CRect(diff.x, 0, 0, 0), point);
				}
				break;
			}
		}

		CView::OnMouseMove(nFlags, point);
		return;
	}

	if (m_bRectPlaced)
	{
		handleIn = (HitTest(point));
		if (handleIn != UNDEFINEDHANDLE)
		{
			SetCursor(cursors[handleIn & ~0x8000]);
		}
		else
		{
			SetCursor(m_OldCursor);
		}

	}

	CView::OnMouseMove(nFlags, point);

}
