// DiffCubeMapDoc.cpp : implementation file
//

#include "stdafx.h"
#include "DCM.h"
#include "DiffCubeMapDoc.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

LPTSTR strDiffuseAppendName[CUBEMAP_TOTAL_DIR] = 
{
	_T("_dpx"),
	_T("_dnx"),
	_T("_dpy"),
	_T("_dny"),
	_T("_dpz"),
	_T("_dnz")
};

/////////////////////////////////////////////////////////////////////////////
// CDiffCubeMapDoc

IMPLEMENT_DYNCREATE(CDiffCubeMapDoc, CDocument)

CDiffCubeMapDoc::CDiffCubeMapDoc()
{
	DWORD i;
	for (i=0; i< CUBEMAP_TOTAL_DIR; i++)
	{
		m_CubeFacesDiffuse[i].m_hObject = NULL;
	}
}

CDiffCubeMapDoc::SetSourceCM(CCubeMapDoc *source)
{
	m_SourceCM = source;
	m_CubeSize = m_SourceCM->m_CubeSize;
}

CDiffCubeMapDoc::~CDiffCubeMapDoc()
{
	DWORD i;

	for (i=0; i< CUBEMAP_TOTAL_DIR; i++)
	{
		if (m_CubeFacesDiffuse[i].m_hObject != NULL)
		{
			m_CubeFacesDiffuse[i].DeleteObject();
		}
	}
}

BEGIN_MESSAGE_MAP(CDiffCubeMapDoc, CDocument)
	//{{AFX_MSG_MAP(CDiffCubeMapDoc)
	ON_COMMAND(ID_FILE_SAVE, OnFileSave)
	ON_COMMAND(ID_FILE_SAVE_AS, OnFileSaveAs)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CDiffCubeMapDoc diagnostics

#ifdef _DEBUG
void CDiffCubeMapDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CDiffCubeMapDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CDiffCubeMapDoc serialization

/////////////////////////////////////////////////////////////////////////////
// CDiffCubeMapDoc commands

BOOL CDiffCubeMapDoc::OnNewDocument() 
{
	D3DXVECTOR3 sampleVec;
	D3DXVECTOR3 *pVec;
	int i;
	RGB *dest;
	int curProgPos;
	CString curLine;
	HBITMAP hBitmapD[CUBEMAP_TOTAL_DIR];
	BITMAPINFO bmi;
	CString mess;
	CPoint pos;
	CCubeMapSize cmSizeDialog;

	if (!CDocument::OnNewDocument())
	{
		return FALSE;
	}

	if (cmSizeDialog.DoModal() == IDCANCEL)
	{
		return FALSE;
	}

	m_CubeSize = cmSizeDialog.m_CubeMapSize;

	bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bmi.bmiHeader.biWidth = m_CubeSize.cx;
	bmi.bmiHeader.biHeight = m_CubeSize.cy;
	bmi.bmiHeader.biPlanes = 1;
	bmi.bmiHeader.biBitCount = 24;
	bmi.bmiHeader.biCompression = BI_RGB;
	bmi.bmiHeader.biSizeImage = m_CubeSize.cx * m_CubeSize.cy;
	bmi.bmiHeader.biXPelsPerMeter = 0;
	bmi.bmiHeader.biYPelsPerMeter = 0;
	bmi.bmiHeader.biClrUsed = 0;
	bmi.bmiHeader.biClrImportant = 0;

	for(i=0; i<CUBEMAP_TOTAL_DIR; i++) 
	{
		hBitmapD[i] = CreateDIBSection(GetDC(NULL), &bmi, DIB_RGB_COLORS, 
							(void**)&m_CubeFacesDiffuseData[i], NULL, 0);
		if (hBitmapD[i] == NULL)
		{
			mess.Format(_T("Unable to create DIB section diffuse cube map face %s.\n"),
							strCubeDir[i]);
			MessageBox(NULL, mess, _T("Cube map allocation error!"), MB_OK);
			return FALSE;
		}
	}

	m_ProgressDialog.Create(CProgressDialog::IDD);
	m_ProgressDialog.m_ProgressCtrl.SetRange(0, 10000);
	m_ProgressDialog.m_ProgressCtrl.SetPos(0);
	m_ProgressDialog.ShowWindow(SW_SHOW);

	for (i=0; i<CUBEMAP_TOTAL_DIR; i++) 
	{
		dest = m_CubeFacesDiffuseData[i];
//		pVec = m_SourceCM->m_CubeNormals[i];
		for (pos.y=0; pos.y<m_CubeSize.cy; pos.y++) 
		{
			curLine.Format(_T("Diffusing cubemap %d line %d"), i, pos.y);
			m_ProgressDialog.m_WorkingString.SetWindowText(curLine);
			m_ProgressDialog.RedrawWindow();

			for (pos.x=0; pos.x<m_CubeSize.cx; pos.x++) 
			{
				// light direction this pixel in cubemap represents
				// sample individual or lookup in table if table present
//				if (m_SourceCM->m_CubeNormals[i] == NULL)
//				{
					m_SourceCM->GetCubeVector(i, &pos, &m_CubeSize, &sampleVec);
					pVec = &sampleVec;
//				}

				DiffuseSample(pVec, dest);
				dest++;
//				pVec++;
				curProgPos = (pos.x * 10000) / m_CubeSize.cx;
				m_ProgressDialog.m_ProgressCtrl.SetPos(curProgPos);
			}
		}
	}

	for (i=0; i< CUBEMAP_TOTAL_DIR; i++)
	{
		m_CubeFacesDiffuse[i].Attach(hBitmapD[i]);
	}

	m_ProgressDialog.DestroyWindow();
	SetModifiedFlag(TRUE);
	return TRUE;
}

void CDiffCubeMapDoc::OnFileSave() 
{
	sprintf(SaveFileName, "%s", (LPCSTR)GetPathName());
	if (!GetSaveFileName(&Sfn))
	{
		return;
	}

	OnSaveDocument(Sfn.lpstrFile);
}

void CDiffCubeMapDoc::OnFileSaveAs()
{
	strcpy(SaveFileName, "");
	if (!GetSaveFileName(&Sfn))
	{
		return;
	}

	OnSaveDocument(Sfn.lpstrFile);
}

BOOL CDiffCubeMapDoc::OnSaveDocument(LPCTSTR lpszPathName) 
{
	DWORD i;
	CString tName;
	CString tExt;
	CString saveName;
	int extPos;

	tName = lpszPathName;
	extPos = tName.ReverseFind('.');
	// strip extension off
	if (extPos == -1)
	{
		tExt = "";
	}
	else
	{
		tExt = tName.Mid(extPos);
		tName = tName.Left(extPos);
	}
	for (i=0; i< CUBEMAP_TOTAL_DIR; i++)
	{
		saveName = tName + strDiffuseAppendName[i] + tExt;
		if (!ilutWinSaveImage(saveName, m_CubeFacesDiffuse[i]))
		{
			tName.Format(_T("Unable to save file %s\nError %d"),
					saveName, GetLastError());
			MessageBox(NULL, tName, _T("Save Error!"), MB_OK);
			return FALSE;
		}
	}
	
	SetModifiedFlag(FALSE);
	return TRUE;
}

//=====================================================================
// Function:	DiffuseSample
// Description:	Sample from all directions facing the normal, except
//				backwards of course
// Parameters:	n = normal direction to sample 
//				colorOut = pointer to RGB output structure
// Returns:		RGB = rgb sum of sample data
//=====================================================================
RGB *CDiffCubeMapDoc::DiffuseSample(const D3DXVECTOR3 *n, RGB *colorOut)
{
	int i;
	D3DXVECTOR3 l;
	D3DXVECTOR3 *pVec;
	float ldotn;
	RGB *src;
	float sum[3];
	float nSamples;
	CPoint pos;
	float divisor;

	nSamples = 0.0f;
	sum[0] = 0.0f;
	sum[1] = 0.0f;
	sum[2] = 0.0f;
	divisor = 0.0f;
	nSamples = 0.0f;
	for(i=0; i<CUBEMAP_TOTAL_DIR; i++) 
	{
		src = m_SourceCM->m_CubeFacesData[i];
		// light direction this pixel in cubemap represents
		pVec = m_SourceCM->m_CubeNormals[i];
		for(pos.y=0; pos.y<m_SourceCM->m_CubeSize.cy; pos.y++) 
		{
			for(pos.x=0; pos.x<m_SourceCM->m_CubeSize.cx; pos.x++) 
			{
				// light direction this pixel in cubemap represents
				if (m_SourceCM->m_CubeNormals[i] == NULL)
				{
					m_SourceCM->GetCubeVector(i, &pos, &m_SourceCM->m_CubeSize, &l);
					pVec = &l;
				}

				ldotn = D3DXVec3Dot(n, pVec);

				if (ldotn >= 0.0f)
				{
					sum[0] += (float)src->R * ldotn;
					sum[1] += (float)src->G * ldotn;
					sum[2] += (float)src->B * ldotn;

					nSamples += 1.0f;
					// for gaussian blur
					divisor += ldotn;
				}
				src++;
				pVec++;
			}
		}
	}


	if (divisor <= 0.0f)
	{
		colorOut->R = 0;
		colorOut->G = 0;
		colorOut->B = 0;
	}
	else
	{
		colorOut->R = (UCHAR)(sum[0] / divisor);
		colorOut->G = (UCHAR)(sum[1] / divisor);
		colorOut->B = (UCHAR)(sum[2] / divisor);
	}

	return colorOut;
}


BOOL CDiffCubeMapDoc::SaveModified() 
{
	CString message;

	if (!IsModified())
		return TRUE;

	if (!theApp.m_GlobalSkipModified)
	{
		message.Format(_T("Document %s is modified. Do you want to save it?\n"),
					(LPCSTR)GetTitle());
		switch (MessageBox(NULL, message, _T("File not saved warning!"), MB_YESNOCANCEL))
		{
		case IDNO:
			return TRUE;
		case IDCANCEL:
			return FALSE;
		}
	}

	strcpy(SaveFileName, "");
	if (!GetSaveFileName(&Sfn))
	{
		return FALSE;
	}

	return OnSaveDocument(Sfn.lpstrFile);
}
